let sync_tbl_sz = 17
let builtins_tbl_sz = 17

open Util

open Format

open Kernel

open Data.Type

open Big_int

module Sync = struct
  type 'a t = {
    mutable signals: 'a list;
    mutable waits: ('a -> unit) list;
    mutex: Mutex.t;
  }

  let create () = {
    signals = [];
    waits = [];
    mutex = Mutex.create ();
  }

  let signal sy v =
    Mutex.lock sy.mutex;
    sy.signals <- v :: sy.signals;
    let waits = sy.waits in
    Mutex.unlock sy.mutex;
    List.iter (fun f -> f v) waits

  let wait sy f =
    Mutex.lock sy.mutex;
    sy.waits <- f :: sy.waits;
    let signals = sy.signals in
    Mutex.unlock sy.mutex;
    List.iter f signals
end

module Node = struct
  type t = {
    depth: int;
    id: int;
    parents: (node * node) option;
    map: int Int.Map.t;
    sz: int;
  }
  and node = {
    node_id: int;
    event: t;
    unlimited: bool;
  }

  exception Incompatible

  let count = Threadref.make 0

  let cumulated_node_comparison_count = Threadref.make 0

  let rec make_map_node event_id n map sz =
    ignore (Threadref.incr cumulated_node_comparison_count);
    match Option.find (Int.Map.find n.node_id) map with
      | None ->
          let map = Int.Map.add n.node_id event_id map in
          make_map_event n.event map (succ sz)
      | Some event_id' ->
          if event_id <> event_id' && not n.unlimited then
            raise Incompatible;
          (map, sz)
  and make_map_event event map sz =
    match event.parents with
      | None -> (map, sz)
      | Some (n0, n1) ->
          let (map, sz) = make_map_node event.id n0 map sz in
          make_map_node event.id n1 map sz

  let make_map_start event_id n0 n1 =
    let map = Int.Map.add n1.node_id event_id n1.event.map in
    make_map_node event_id n0 map (succ n1.event.sz)

  let make_map event_id parents =
    match parents with
      | None -> (Int.Map.empty, 0, 0)
      | Some (n0, n1) ->
          let (map, sz) =
            if n0.event.sz <= n1.event.sz then
              make_map_start event_id n0 n1
            else
              make_map_start event_id n1 n0 in
          (map, sz, max n0.event.depth n1.event.depth)

  let make parents =
    let id = Threadref.incr count in
    let (map, sz, depth) = make_map id parents in
    { id = id; depth = depth; parents = parents; map = map; sz = sz }

  let merge_count = Threadref.make 0

  let incompatible_merge_count = Threadref.make 0

  let cumulated_merge_time = Threadref.make 0.0

  let try_make parents =
    ignore (Threadref.incr merge_count);
    let time = Sys.time () in
    try
      let result = make parents in
      let _ =
        Threadref.apply ((+.) (Sys.time () -. time)) cumulated_merge_time in
      Some result
    with Incompatible ->
      ignore (Threadref.incr incompatible_merge_count);
      let _ =
        Threadref.apply ((+.) (Sys.time () -. time)) cumulated_merge_time in
      None

  let node_count = Threadref.make 0

  let make_node event unlimited =
    let node_id = Threadref.incr node_count in
    { node_id = node_id; event = event; unlimited = unlimited; }

  let rec restore_rec event map depths depth =
    match event.parents, depths with
      | _, [] -> (map, event)
      | None, _ -> failwith "restore_rec"
      | Some (left, right), hd :: tl ->
          if hd = depth then
            let map = Int.Map.remove left.node_id map in
            restore_rec right.event map tl (succ depth)
          else
            let (map, right_event) =
              restore_rec right.event map depths (succ depth) in
            let parents = Some (left, { right with event = right_event }) in
            let event = { event with parents = parents } in
            (map, event)

  let rec restore event depths =
    let (map, event) = restore_rec event event.map depths 0 in
    { event with map = map }
      

  let reset_depth event = { event with depth = 0 }

  let incr_depth event = { event with depth = succ event.depth }
end

module Tok = struct
  type t = {
    node: Node.node;
    contents: Data.t array
  }

  let make event contents =
    let node = Node.make_node event false in
    { node = node; contents = contents }
end

module V = struct
  type t = {
    addr: int;
    tbl: Tok.t Sync.t Sgn.Hashtbl.t;
  }

  let count = Threadref.make 0

  let fresh () =
    let addr = Threadref.incr count in
    let tbl = Sgn.Hashtbl.create sync_tbl_sz in
    { addr = addr; tbl = tbl }

  let sync var sgn =
    try Sgn.Hashtbl.find var.tbl sgn
    with Not_found ->
      let sy = Sync.create () in
      Sgn.Hashtbl.add var.tbl sgn sy;
      sy

  let format fmt var = fprintf fmt "$%i" var.addr
end

let var_ty = Data.new_type V.format

let pp_print_big_int fmt i = fprintf fmt "%s" (string_of_big_int i)

let big_int_ty = Data.new_type pp_print_big_int

let int_inj i = big_int_ty.inj (big_int_of_int i)

let int_proj d = int_of_big_int (big_int_ty.proj d)

let pp_print_float fmt f = fprintf fmt "%f" f

let float_ty = Data.new_type pp_print_float

let pp_print_mem fmt mem = fprintf fmt "<mem>"

let mem_ty = Data.new_type pp_print_mem

let pp_print_array fmt array =
  fprintf fmt "{%a}" (Array.format "," Data.format) array

let array_ty = Data.new_type pp_print_array

let string_inj s =
  mem_ty.inj (s ^ String.make 1 '\000', 0)

let string_proj v =
  let (s, off) = mem_ty.proj v in
  let l =
    try String.index_from s off '\000' - off
    with Not_found ->
      prerr_endline "Illformed string.\n";
      exit 1 in
  String.sub s off l

let list_inj l = array_ty.inj (Array.of_list l)

let list_proj v = Array.to_list (array_ty.proj v)

let bool_inj b = if b then int_inj 1 else int_inj 0

let bool_proj v = not (eq_big_int (big_int_ty.proj v) zero_big_int)

let char_inj c = int_inj (int_of_char c)

let char_proj v = char_of_int (int_proj v)

let lookup ctx var =
  match var with
    | Am.Local index -> ctx.Am.local_vars.(index)
    | Am.Param index -> ctx.Am.param_vars.(index)
    | Am.Env index -> ctx.Am.env_vars.(index)
    | Am.Global index -> ctx.Am.global_vars.(index)
    | Am.Ask_var -> ctx.Am.ask_var

let rec eval ctx v =
  match v with
    | Am.Var x -> lookup ctx x
    | Am.Term (Term.Const (Const.Int i)) -> big_int_ty.inj i
    | Am.Term (Term.Const (Const.Float f)) -> float_ty.inj f
    | Am.Term (Term.Const (Const.String s)) -> string_inj s
    | Am.Term (Term.Array a) -> array_ty.inj (Array.map (eval ctx) a)

let tell' event sy a =
  let tok = Tok.make event a in
  Sync.signal sy tok;
  tok

let tell event m t =
  let sy = V.sync (var_ty.proj m) t.Compound.s in
  tell' event sy t.Compound.args

let tell_trace' event sy v f p a =
  Trace.print p "tell" "$%d.%s(%a)"
    v.V.addr f
    (Array.format "," Data.format) a;
  ignore (tell' event sy a)

let tell_trace event p m t =
  let v = var_ty.proj m in
  let sy = V.sync v t.Compound.s in
  tell_trace' event sy v (Compound.f t) p t.Compound.args

let ask' event sy k =
  let node = Node.make_node event true in
  Sync.wait sy begin fun tok ->
    match Node.try_make (Some (tok.Tok.node, node)) with
      | None -> ()
      | Some event -> k event tok
  end

let ask event m s k =
  let sy = V.sync (var_ty.proj m) s in
  ask' event sy k

let restore event p (m, t) =
   tell_trace event p m t

let void_builtins_tbl = Sgn.Hashtbl.create builtins_tbl_sz

let value_builtins_tbl = Sgn.Hashtbl.create builtins_tbl_sz

let bool_builtins_tbl = Sgn.Hashtbl.create builtins_tbl_sz

let decl_void_builtin spec s f =
  let b = {
    Opt.Builtin.Void.desc = { Opt.Builtin.s = s };
    spec = spec;
  } in
  Sgn.Hashtbl.add Opt.builtins_tbl s (Opt.Builtin.Void b);
  Sgn.Hashtbl.add void_builtins_tbl s f

let decl_simple_builtin = decl_void_builtin Opt.Builtin.Void.Simple

let decl_do_builtin = decl_void_builtin Opt.Builtin.Void.Do

let decl_value_builtin s f =
  let b = {
    Opt.Builtin.Value.desc = { Opt.Builtin.s = s };
  } in
  Sgn.Hashtbl.add Opt.builtins_tbl s (Opt.Builtin.Value b);
  Sgn.Hashtbl.add value_builtins_tbl s f

let decl_bool_builtin s f =
  let b = {
    Opt.Builtin.Bool.desc = { Opt.Builtin.s = s };
  } in
  Sgn.Hashtbl.add Opt.builtins_tbl s (Opt.Builtin.Bool b);
  Sgn.Hashtbl.add bool_builtins_tbl s f

let do_call_void event ctx p s args =
  let args = Array.map (eval ctx) args in
  let f = Sgn.Hashtbl.find void_builtins_tbl s in
  Trace.print p "call" "%s(%a)" s.Sgn.f (Array.format "," Data.format) args;
  f event args

let do_call_value event ctx p s args result =
  let args = Array.map (eval ctx) args in
  let f = Sgn.Hashtbl.find value_builtins_tbl s in
  Trace.print p "call" "?? <- %s(%a)"
    s.Sgn.f (Array.format "," Data.format) args;
  let r = f event args in
  Trace.print p "return" "%a" Data.format r;
  ctx.Am.local_vars.(result) <- r

let call event ctx (p, c) =
  match c with
    | Am.Void (s, args) ->
        do_call_void event ctx p s args
    | Am.Value (s, args, result) ->
        do_call_value event ctx p s args result

let bool_call execute_proc execute_code event ctx (p, s, a, code_t, code_f) =
  let args = Array.map (eval ctx) a in
  let f = Sgn.Hashtbl.find bool_builtins_tbl s in
  Trace.print p "call" "if %s(%a)"
    s.Sgn.f (Array.format "," Data.format) args;
  if f event args then
    begin
      Trace.print p "return" "true";
      execute_code execute_proc event ctx code_t
    end
  else
    begin
      Trace.print p "return" "false";
      execute_code execute_proc event ctx code_f
    end

let do_tell event ctx sy v s (p, a) =
  tell_trace' event sy v s.Sgn.f p (Array.map (eval ctx) a)

let fire_count = Threadref.make 0

let pp_tok fmt _ = pp_print_string fmt "<tok>"

type tok_info = {
  mutable restored: bool;
  v: V.t;
  s: Sgn.t;
  contents: Data.t array;
}

let tok_ty = Data.new_type pp_tok

let pp_print_dummy fmt () = fprintf fmt "<dummy>"

let dummy_ty = Data.new_type pp_print_dummy

let dummy = dummy_ty.inj ()

let fire execute_proc p id event ask_var env params =
  ignore (Threadref.incr fire_count);
  Trace.print p "fire" "[%d(%a)](%a)"
    id (Array.format "," Data.format) env
    (Array.format "," Data.format) params;
  execute_proc event id ask_var env params

let do_ask execute_proc event ctx sy v s agent_call =
  let p = agent_call.Am.pos in
  let f = s.Sgn.f in
  let id = agent_call.Am.id in
  let env = Array.map (lookup ctx) agent_call.Am.env in
  Trace.print p "ask" "$%d.%s [%d(%a)]"
    v.V.addr f
    id (Array.format "," Data.format) env;
  ask' event sy begin fun event tok ->
    let tok_info = {
      restored = false;
      v = v;
      s = s;
      contents = tok.Tok.contents;
    } in
    let ask_var = tok_ty.inj tok_info in
    fire execute_proc p id event ask_var env tok.Tok.contents
  end

let restore_mutex = Mutex.create ()

let restore event ctx restores =
  let check (tells, to_restore) (v, depth) =
  let v = tok_ty.proj (lookup ctx v) in
  Mutex.lock restore_mutex;
  if v.restored then
    begin
      Mutex.unlock restore_mutex;
      ((v.v, v.s, v.contents) :: tells, to_restore)
    end
  else
    begin
      v.restored <- true;
      Mutex.unlock restore_mutex;
      (tells, depth :: to_restore)
    end in
  let (tells, to_restore) = List.fold_left check ([], []) restores in
  let to_restore = List.rev to_restore in
  let event = Node.restore event to_restore in
  List.iter
    (fun (v, s, contents) ->
       tell_trace' event (V.sync v s) v s.Sgn.f Pos.dummy_range contents) tells;
  event

let tell_call execute_proc event ctx (a, agent_call) =
  let a = Array.map (eval ctx) a in
  let env = Array.map (lookup ctx) agent_call.Am.env in
  fire execute_proc agent_call.Am.pos agent_call.Am.id event dummy env a

let ask_tell execute_proc event ctx v ask_tell =
  let s = ask_tell.Am.s in
  let sy = V.sync v s in
  List.iter (do_ask execute_proc event ctx sy v s) ask_tell.Am.asks;
  List.iter (do_tell event ctx sy v s) ask_tell.Am.tells

let ask_tells execute_proc event ctx (x, list) =
  let v = var_ty.proj (lookup ctx x) in
  List.iter (ask_tell execute_proc event ctx v) list

let rec execute_code execute_proc event ctx code =
  let init i = ctx.Am.local_vars.(i) <- var_ty.inj (V.fresh ()) in
  List.iter init code.Am.init;
  let event = restore event ctx code.Am.restores in
  List.iter (call event ctx) code.Am.calls;
  List.iter (tell_call execute_proc event ctx) code.Am.tell_calls;
  List.iter (ask_tells execute_proc event ctx) code.Am.ask_tells;
  List.iter (bool_call execute_proc execute_code event ctx) code.Am.bool_calls

let max_depth = ref 1000

let set_max_depth new_max_depth =
  max_depth := int_of_string new_max_depth

let () =
  Cmdline.register ~long:"max-depth"
    (Cmdline.One set_max_depth)
    ~help:"sets maximum depth-first steps before breadth exploration"

let max_thread = ref 10

let set_max_thread new_max_thread =
  max_thread := int_of_string new_max_thread

let () =
  Cmdline.register ~long:"max-thread"
    (Cmdline.One set_max_thread)
    ~help:"sets maximum threads count"

let threads = ref []

let thread_count = ref 0

let job_queue = Queue.create ()

let job_queue_mutex = Mutex.create ()

let rec perform f =
  f ();
  Mutex.lock job_queue_mutex;
  let next =
    try Some (Queue.pop job_queue)
    with Queue.Empty -> None in
  Mutex.unlock job_queue_mutex;
  match next with
    | None -> decr thread_count
    | Some f -> perform f

let add_thread f =
  Mutex.lock job_queue_mutex;
  if thread_count < max_thread then
    begin
      incr thread_count;
      threads := Thread.create perform f :: !threads
    end
  else
    Queue.add f job_queue;
  Mutex.unlock job_queue_mutex

let execute_proc' global_vars execute_proc event proc ask_var env_vars param_vars
    () =
  let local_vars = Array.make proc.Am.local_count dummy in
  let ctx = {
    Am.local_vars = local_vars;
    param_vars = param_vars;
    env_vars = env_vars;
    global_vars = global_vars;
    ask_var = ask_var;
  } in
  execute_code execute_proc event ctx proc.Am.code

let rec execute_proc global_vars code_tbl event id ask_var env_vars param_vars =
  let proc = code_tbl.(id) in
  if event.Node.depth <= !max_depth then
    let event = Node.incr_depth event in
    execute_proc' global_vars (execute_proc global_vars code_tbl) event proc
      ask_var env_vars param_vars ()
  else
    let event = Node.reset_depth event in
    add_thread
      (execute_proc' global_vars (execute_proc global_vars code_tbl) event proc
         ask_var env_vars param_vars)

let start code_tbl global_vars =
  let proc = code_tbl.(0) in
  let event = Node.make None in
  let ctx = {
    Am.local_vars = global_vars;
    param_vars = [| |];
    env_vars = [| |];
    global_vars = global_vars;
    ask_var = dummy;
  } in
  execute_code (execute_proc global_vars code_tbl) event ctx proc.Am.code

let error_on_exhausted = ref false

let enable_error_on_exhausted () =
  error_on_exhausted := true

let () =
  Cmdline.register ~long:"error-on-exhausted"
    (Cmdline.Zero enable_error_on_exhausted)
    ~help:"fails if there is no more asks to fire"

let rec wait_threads () =
  Mutex.lock job_queue_mutex;
  match !threads with
    | [] -> Mutex.unlock job_queue_mutex
    | hd :: tl ->
        threads := tl;
        Mutex.unlock job_queue_mutex;
        Thread.join hd;
        wait_threads ()

let execute code_tbl global_vars =
  incr thread_count;
  let () = start code_tbl global_vars in
  perform (fun () -> ());
  wait_threads ();
  if !error_on_exhausted then
    Err.exhausted ()

let print_statistics fmt =
  let cumulated_node_comparison_count =
    Threadref.get Node.cumulated_node_comparison_count in
  let merge_count = Threadref.get Node.merge_count in
  eprintf "\
Fire count: %d
Cumulated merge time: %f
Merge count: %d
Incompatible merge count: %d
Cumulated node comparison count: %d (%f per merge)@."
    (Threadref.get fire_count)
    (Threadref.get Node.cumulated_merge_time)
    merge_count
    (Threadref.get Node.incompatible_merge_count)
    cumulated_node_comparison_count
    (float_of_int cumulated_node_comparison_count /. float_of_int merge_count)
